def retrieve_top_hyper(study):
    """
    Truy xuất các hyperparameter hàng đầu từ một đối tượng Optuna Study.

    Args:
    study (optuna.study.Study): Đối tượng Optuna Study.

    Returns:
    list: Danh sách chứa các hyperparameter hàng đầu.
    """
    # Truy xuất tất cả các thử nghiệm
    trials = study.trials

    # Sắp xếp các thử nghiệm dựa trên các giá trị mục tiêu
    trials.sort(key=lambda trial: trial.values, reverse=True)

    # Lấy ra 100 thử nghiệm hàng đầu
    top_trials = trials[:100]

    # Trích xuất các hyperparameter từ các thử nghiệm hàng đầu
    top_hyperparameters = [trial.params for trial in top_trials]

    return top_hyperparameters



import numpy as np
import pandas as pd

from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split

def split_optuna_data(data):
    """
    Hàm này chia dữ liệu thành các tập train và test zđể sử dụng trong quá trình tối ưu hóa bằng Optuna.

    Args:
    data (pandas.DataFrame): DataFrame chứa dữ liệu cần chia.

    Returns:
    pandas.DataFrame: DataFrame chứa dữ liệu train (đã được chuẩn hóa).
    pandas.DataFrame: DataFrame chứa dữ liệu test (đã được chuẩn hóa).
    pandas.Series: Series chứa nhãn tương ứng với dữ liệu train.
    pandas.Series: Series chứa nhãn tương ứng với dữ liệu test.
    """
    new_part = np.array_split(data, 3)

    # Truy cập từng phần một
    hold_out = new_part[2]
    train_data = pd.concat([new_part[0], new_part[1]], axis=0)
    
    # Loại bỏ các cột không cần thiết
    if 'open' in train_data.columns:
        optuna_data = train_data.drop(['close', 'open', 'high', 'low', 'volume', 'Changes'], axis=1)
    else:
        optuna_data = train_data.drop(['close', 'Changes'], axis=1)
    
    scaler = StandardScaler()
    scaler.fit(optuna_data)
    optuna_data = pd.DataFrame(scaler.transform(optuna_data))    
    
    # Chia dữ liệu thành tập train và tập test
    X_train, X_valid, y_train, y_valid = train_test_split(optuna_data, train_data['Changes'], test_size=0.5, shuffle=False)

    return X_train, X_valid, y_train, y_valid


import requests
import pandas as pd
from datetime import datetime, timedelta

def get_vn30f():
    """
    Hàm này trả về dữ liệu giá cổ phiếu VN30F từ API

    Parameters
    ----------
    None
    
    Returns
    --------
    vn30fm : pandas.DataFrame
        DataFrame chứa dữ liệu giá cổ phiếu VN30F bao gồm cột 'Date' là thời gian, 
        'open' là giá mở cửa, 'high' là giá cao nhất, 'low' là giá thấp nhất, 'close' là giá đóng cửa, 
        và 'volume' là khối lượng giao dịch.
    """
    def vn30f():
        """
        Hàm này truy vấn API để lấy dữ liệu giá cổ phiếu VN30F từ API.

        Returns:
        pandas.DataFrame: Dữ liệu giá cổ phiếu VN30F dưới dạng pandas dataframe
        """
        return requests.get("https://services.entrade.com.vn/chart-api/chart?from=0000000000&resolution=1&symbol=VN30F1M&to=9999999999").json()

    # Lấy dữ liệu từ API và chuyển đổi thành DataFrame
    vn30fm = pd.DataFrame(vn30f()).iloc[:, :6]

    # Chuyển đổi cột 't' sang định dạng datetime và cập nhật múi giờ
    vn30fm['t'] = vn30fm['t'].astype(int).apply(lambda x: datetime.utcfromtimestamp(x) + timedelta(hours=7))

    # Đổi tên cột
    vn30fm.columns = ['Date', 'open', 'high', 'low', 'close', 'volume']

    # Tạo dictionary chứa các hàm tổng hợp cho các cột giá
    ohlc_dict = {
        'open': 'first',
        'high': 'max',
        'low': 'min',
        'close': 'last',
        'volume': 'sum',
    }

    # Đưa dữ liệu về định dạng chuẩn OHLC (Open, High, Low, Close)
    vn30fm = vn30fm.set_index('Date').resample('1D').apply(ohlc_dict)

    return vn30fm


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

def choose_position(change):
    """
    Identify position according to value of change.
    
    This function will take change and decide the position. 
    If change is larger than 0, position will be long position,
    if change is smaller than 0, position will be short position,
    otherwise position will be neutral position.

    Parameters
    ----------
    change : numpy.float32
    
    Returns
    -------
    pos : int
    """
    if change >0:
        pos = 1
    elif change == 0:
        pos = 0
    else:
        pos = -1
    
    return pos

def backtest_position_ps(position, change): 
    """
    Parameters
    ----------
    position : 
    change :
    
    Returns
    -------
    pnk_long_short : 
    """
    pos = pd.Series(position).shift(1).fillna(0)
    ch = pd.Series(change).shift(1).fillna(0)

    pos_long = np.where(pos > 0, pos, 0)
    pos_short = np.where(pos < 0, pos, 0)
    pnl_long_short = ch * pos_long + ch * pos_short

    return pnl_long_short

def draw_corr(corr):
    """
    Vẽ biểu đồ ma trận tương quan.

    Args:
    corr (numpy.ndarray or pandas.DataFrame): Ma trận tương quan.

    Returns:
    None
    """
    plt.matshow(corr)  # Invert y-axis to get origo at lower left corner
    plt.gca().xaxis.tick_bottom()
    plt.gca().invert_yaxis()
    plt.colorbar()
    plt.show()
    
def cov2corr(cov):
    """
    Chuyển ma trận hiệp phương sai thành ma trận tương quan tương ứng.

    Args:
    cov (numpy.ndarray): Ma trận hiệp phương sai.

    Returns:
    numpy.ndarray: Ma trận tương quan tương ứng.
    """
    # Tính độ lệch chuẩn của từng biến
    std = np.sqrt(np.diag(cov))
    
    # Tính ma trận tương quan từ ma trận hiệp phương sai
    corr = cov / np.outer(std, std)
    
    # Xử lý các giá trị ngoại lai gây ra bởi lỗi số học
    corr[corr < -1] = -1
    corr[corr > 1] = 1
    
    return corr



__all__ = ['data', 'evaluate', 'ocn', 'optuna', 'choose_position', 'backtest_position_ps', 'draw_corr', 'cov2corr']

from . import data, evaluate, ocn, optuna
from .utils import choose_position, backtest_position_ps, draw_corr, cov2corr



import numpy as np
import pandas as pd

def remove_outliers(df, threshold=90):
    """
    Remove outliers from dataset.

    Parameters
    ----------
    df : pandas.DataFrame 
        DataFrame chứa dữ liệu cần xử lý, có cột 'close' chứa giá đóng cửa.
    threshold : float (default=90)
        Ngưỡng phần trăm để xác định giá trị outlier. Mặc định là 90.

    Returns
    -------
    df : pandas.DataFrame
        DataFrame đã được xử lý để loại bỏ các giá trị outlier.
    """
    # Tính toán phần trăm thay đổi giá đóng cửa
    percentage_change = df['close'].pct_change() * 100

    # Tạo mask để xác định các giá trị outlier dựa trên ngưỡng threshold
    mask = (percentage_change > threshold) & (percentage_change < -threshold)

    # Thiết lập giá trị thay đổi dựa trên phần trăm thay đổi âm và dương
    change = df['close'].shift(-1) * (0.991)
    change[mask] = df['close'].shift(-1) * 1.009

    # Áp dụng giá trị thay đổi cho các giá trị outlier
    df.loc[mask, 'close'] = df.loc[mask, 'close'].shift(-1) * change.shift(-1)

    return df

def add_index(data):
    """
    Hàm này thêm các index mới cho dữ liệu.

    Args:
    data (pandas.DataFrame): DataFrame chứa dữ liệu cần thêm index.

    Returns:
    pandas.DataFrame: DataFrame đã được mở rộng với các index mới.
    """
    # Chia cột index thành cột 'time' và 'Date'
    data['time'] = [str(i)[11:] for i in data.index]
    data['Date'] = [str(i)[:10] for i in data.index]

    # Tạo bảng pivot với các cột 'open', 'high', 'low', 'close', 'volume' và giá trị là giá trị tương ứng của chúng
    pivot = data.pivot(index='Date', columns='time', values=['open', 'high', 'low', 'close', 'volume'])

    # Điền các giá trị NaN bằng phương pháp 'bfill' và stack lại dữ liệu
    fill = pivot.fillna(method='bfill').stack('time')

    return fill




__all__ = ['get_vn30f', 'remove_outliers', 'add_index', 'split_optuna_data']

from .get_data import get_vn30f
from .preprocess import remove_outliers, add_index
from .split_dataset import split_optuna_data
from .utils import *


import numpy as np
import pandas as pd
import xgboost as xgb

from ..utils import choose_position, backtest_position_ps

def calculate_sharpe_ratio(pnl):
    """
    Calculate Sharpe ratio given PNL.
    
    Sharpe ratio illustrates of expected profit for each unit of risk in specific time interval
    and will be calculated by this formula: 
    = (Average of PNL values / Standard deviation of PNL values) * sqrt(252)
    252 is the number of trading days in a year, 
    we multiply with square root of 252 to convert sharpe ratio daily to sharpe ratio annually.
    
    Parameters
    ----------
    pnl : pandas.core.series.Series
        The Series of PNL values during the time interval (252).
    
    Returns
    -------
    sharpe : numpy.float64
        The sharpe ratio during the time interval.
    """
    avg_return = np.mean(pnl)
    std_return = np.std(pnl)
    sharpe = (avg_return / std_return)* np.sqrt(len(pnl))
    return sharpe

def sharpe_for_vn30f(y_pred, y_groud):
    """
    Tính toán tỷ lệ Sharpe cho một mô hình dự đoán giá VN30F.

    Args:
    y_pred (numpy.ndarray): Mảng chứa dự đoán của mô hình.
    y_groud (numpy.ndarray): Mảng chứa giá trị thực tế của VN30F.

    Returns:
    numpy.ndarray: Mảng chứa các vị trí được chọn dựa trên dự đoán của mô hình.
    pandas.Series: Chuỗi chứa PNL tích lũy.
    pandas.Series: Chuỗi chứa PNL hàng ngày được tích lũy.
    float: Tỷ lệ Sharpe của PNL hàng ngày.
    """
    # Chọn vị trí dựa trên dự đoán của mô hình
    pos = [choose_position(change) for change in y_pred]
    pos = np.array(pos)
    
    # Tính toán PNL hàng ngày
    pnl = backtest_position_ps(pos, y_groud)
    pnl = np.cumsum(pnl)
    cumulative_sum = pnl.groupby(level=0).last().values
    unique_dates = pnl.groupby(level=0).last().index

    # Tạo một Series mới với các giá trị của cumulative_sum và unique_dates
    daily_pnl = pd.Series(cumulative_sum, index=unique_dates)
    
    # Tính toán tỷ lệ Sharpe
    sharpe = calculate_sharpe_ratio(daily_pnl)

    return pos, pnl, daily_pnl, sharpe

def retrieve_top_pnl(top_hyperparameters, X_train, y_train, X_valid, y_valid):
    """
    Truy xuất PNL hàng đầu từ các siêu tham số và dữ liệu huấn luyện.

    Args:
    top_hyperparameters (list): Danh sách chứa các siêu tham số hàng đầu.
    X_train (numpy.ndarray or pandas.DataFrame): Dữ liệu huấn luyện.
    y_train (numpy.ndarray or pandas.Series): Nhãn của dữ liệu huấn luyện.
    X_valid (numpy.ndarray or pandas.DataFrame): Dữ liệu kiểm tra.
    y_valid (numpy.ndarray or pandas.Series): Nhãn của dữ liệu kiểm tra.

    Returns:
    list: Danh sách chứa PNL của các siêu tham số hàng đầu.
    """
    top_pnl = []
    
    for params in top_hyperparameters:
        # Tạo và huấn luyện mô hình
        model = xgb.XGBRegressor(**params)
        model.fit(X_train, y_train)
        
        # Đưa ra dự đoán
        y_pred_valid = model.predict(X_valid)
        _, pnl_valid, _, _ = sharpe_for_vn30f(y_pred_valid, y_valid)
        pnl_valid_no_nan = np.nan_to_num(pnl_valid, nan=0)
        top_pnl.append(pnl_valid_no_nan)

    return top_pnl




from setuptools import find_packages, setup

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name="demo_alphav2",
    version="2.0.1",
    description="A package of alphav2",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/quandn2003/demo_package_alphav2",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    # install_requires=[
    #     'matplotlib==3.7.4',
    #     'numpy==1.23.5',
    #     'pandas==2.0.3',
    #     'Requests==2.31.0',
    #     'scikit_learn==1.3.2',
    #     'setuptools==69.0.3',
    #     'xgboost==2.0.3'
    # ],
    # extras_require={
    #     'dev': ['pytest', 'sphinx'],
    #     # 'gpu': ['cupy', 'tensorflow-gpu']
    # },
    python_requires=">=3.10",
)


import xgboost as xgb
import numpy as np

from ..evaluate.evaluate import sharpe_for_vn30f

def objective_feature_vn30f(trial, X_train, X_valid, y_train, y_valid):
    """
    Return values of metrics of each trial object to find set of features for training XGBoost model.
    
    This function will create a search space that contain binary bits for each feature in input dataset. 
    The 1 bit representing for selecting that feature, otherwise it's bit 0. 
    Then the XGBoost model is trained by using selected features and evaluated by two metrics:
    - sharpe_oos: Sharpe ratio calculated on the out-of-sample (validation) dataset.
    - generalization_score: Absolute ratio of out-of-samples sharpe ratio to in-sample sharpe ratio)    
    
    Parameters
    ----------
    trial : Trial
        Optuna `Trial` object used for hyperparameter optimization.
    X_train : pd.DataFrame
        Pandas DataFrame containing features of the training dataset.
    X_valid : pd.DataFrame
        Pandas DataFrame containing features of the validation dataset.
    y_train : pd.DataFrame
        Pandas DataFrame containing target variables of the training dataset.
    y_valid : pd.DataFrame
        Pandas DataFrame containing target variables of the validation dataset.
    
    Returns
    -------
    sharpe_oos : numpy.float64
        Sharpe ratio calculated on the out-of-sample (validation) dataset.
    generalization_score : numpy.float64
        Absolute ratio of out-of-sample Sharpe ratio to in-sample Sharpe ratio.
    """
    selected_features = []
    for col in X_train.columns:
        if trial.suggest_categorical(col, [0, 1]):
            selected_features.append(col)

    X_train_selected = X_train[selected_features]
    X_valid_selected = X_valid[selected_features]

    model = xgb.XGBRegressor()
    model.fit(X_train_selected, y_train)

    y_pred_train = model.predict(X_train_selected)
    y_pred_valid = model.predict(X_valid_selected)

    _, _, _, sharpe_is = sharpe_for_vn30f(y_pred_train, y_train)
    _, _, _, sharpe_oos = sharpe_for_vn30f(y_pred_valid, y_valid)
    
    generalization_ratio = abs(sharpe_oos / sharpe_is)

    return sharpe_oos, generalization_ratio

def objective_model_vn30f(trial, X_train, X_valid, y_train, y_valid):
    """
    Return values of metrics of each trial object to find the optimized set of hyperparameters for XGBoost model.
    
    This function will create a search space for each hyperparameters in XGBoost model. 
    The 1 bit representing for selecting that feature, otherwise it's bit 0. 
    Then the XGBoost model is trained by using optimized hyperparameters and evaluated by metric:
    - sharpe_oos: Sharpe ratio calculated on the out-of-sample (validation) dataset.
    
    Parameters
    ----------
    trial : Trial
        Optuna `Trial` object used for hyperparameter optimization.
    X_train : pd.DataFrame
        Pandas DataFrame containing features of the training dataset.
    X_valid : pd.DataFrame
        Pandas DataFrame containing features of the validation dataset.
    y_train : pd.DataFrame
        Pandas DataFrame containing target variables of the training dataset.
    y_valid : pd.DataFrame
        Pandas DataFrame containing target variables of the validation dataset.
    
    Returns
    -------
    sharpe_oos : numpy.float64
        Sharpe ratio calculated on the out-of-sample (validation) dataset.
    generalization_score : numpy.float64
        Absolute ratio of out-of-sample Sharpe ratio to in-sample Sharpe ratio.
    """
    params = {
        'max_depth': trial.suggest_int('max_depth', 1, 10),
        'learning_rate': trial.suggest_float('learning_rate', 0.01, 1.0),
        'n_estimators': 200,
        'min_child_weight': trial.suggest_int('min_child_weight', 1, 10),
        'gamma': trial.suggest_float('gamma', 0.01, 1.0),
        'subsample': trial.suggest_float('subsample', 0.01, 1.0),
        'colsample_bytree': trial.suggest_float('colsample_bytree', 0.01, 1.0),
        'reg_alpha': trial.suggest_float('reg_alpha', 0.01, 1.0),
        'reg_lambda': trial.suggest_float('reg_lambda', 0.01, 1.0),
        'random_state': trial.suggest_int('random_state', 1, 1000)
    }

    model = xgb.XGBRegressor(**params)
    model.fit(X_train, y_train)
    
    y_pred_train = model.predict(X_train)
    y_pred_valid = model.predict(X_valid)
    
    _, _, _, sharpe_is = sharpe_for_vn30f(y_pred_train, y_train)
    _, _, _, sharpe_oos = sharpe_for_vn30f(y_pred_valid, y_valid)

    generalization_ratio = abs(sharpe_oos / sharpe_is)

    return sharpe_oos, generalization_ratio


def clusterKMeansBase(corr0, maxNumClusters=10, n_init=10, debug=False):
    """
    Phân cụm các biến tài sản dựa trên ma trận tương quan sử dụng thuật toán KMeans.

    Args:
    corr0 (numpy.ndarray): Ma trận tương quan.
    maxNumClusters (int, optional): Số lượng cụm tối đa. Mặc định là 10.
    n_init (int, optional): Số lần khởi tạo ngẫu nhiên. Mặc định là 10.
    debug (bool, optional): Chế độ debug. Mặc định là False.

    Returns:
    numpy.ndarray: Ma trận tương quan đã được sắp xếp lại.
    dict: Từ điển chứa thông tin về các cụm.
    pandas.Series: Series chứa hệ số Silhouette tối ưu cho từng quan sát.
    """
    # Xử lý các giá trị ngoại lai trong ma trận tương quan
    corr0[corr0 > 1] = 1
    
    # Tính ma trận khoảng cách từ ma trận tương quan
    dist_matrix = ((1 - corr0) / 2.) ** 0.5
    
    # Tạo một Series trống để lưu trữ hệ số Silhouette tối ưu
    silh_coef_optimal = pd.Series(dtype='float64')
    
    # Xác định số cụm tối đa
    maxNumClusters = min(maxNumClusters, int(np.floor(dist_matrix.shape[0] / 2)))
    
    # Lặp qua các lần khởi tạo ngẫu nhiên
    for init in range(0, n_init):
        # Lặp qua số lượng cụm từ 2 đến maxNumClusters
        for num_clusters in range(2, maxNumClusters + 1):
            kmeans_ = KMeans(n_clusters=num_clusters, n_init=10)
            kmeans_ = kmeans_.fit(dist_matrix)
            silh_coef = silhouette_samples(dist_matrix, kmeans_.labels_)
            stat = (silh_coef.mean() / silh_coef.std(), silh_coef_optimal.mean() / silh_coef_optimal.std())

            # Nếu hệ số này tốt hơn hệ số tối ưu trước đó, thiết lập là tối ưu
            if np.isnan(stat[1]) or stat[0] > stat[1]:
                silh_coef_optimal = silh_coef
                kmeans = kmeans_
                if debug:
                    print(kmeans)
                    print(stat)
                    silhouette_avg = silhouette_score(dist_matrix, kmeans_.labels_)
                    print("For n_clusters = " + str(num_clusters) + " The average silhouette_score is : " + str(silhouette_avg))
                    print("********")
    
    # Sắp xếp lại ma trận tương quan theo nhãn cụm
    newIdx = np.argsort(kmeans.labels_)
    corr1 = corr0.iloc[newIdx]
    corr1 = corr1.iloc[:, newIdx]

    # Tạo từ điển chứa thông tin về các cụm
    clstrs = {i: corr0.columns[np.where(kmeans.labels_ == i)[0]].tolist() for i in np.unique(kmeans.labels_)}
    
    # Tạo Series chứa hệ số Silhouette tối ưu
    silh_coef_optimal = pd.Series(silh_coef_optimal, index=dist_matrix.index)
    
    return corr1, clstrs, silh_coef_optimal

''' Improve number clusters using silh scores

    :param corr_mat: (pd.DataFrame) Correlation matrix
    :param clusters: (dict) Clusters elements
    :param top_clusters: (dict) Improved clusters elements
    :return: (tuple) [ordered correlation matrix, clusters, silh scores]
'''
def makeNewOutputs(corr0, clstrs, clstrs2):
    """
    Tạo ra các đầu ra mới từ các thông tin về các cụm.

    Args:
    corr0 (numpy.ndarray or pandas.DataFrame): Ma trận tương quan gốc.
    clstrs (dict): Từ điển chứa thông tin về các cụm.
    clstrs2 (dict): Từ điển chứa thông tin về các cụm mới.

    Returns:
    numpy.ndarray or pandas.DataFrame: Ma trận tương quan mới.
    dict: Từ điển chứa thông tin về các cụm mới.
    pandas.Series: Series chứa hệ số Silhouette mới cho từng quan sát.
    """
    # Khởi tạo từ điển và danh sách mới
    clstrsNew, newIdx = {}, []
    
    # Tạo từ điển mới từ thông tin về các cụm gốc
    for i in clstrs.keys():
        clstrsNew[len(clstrsNew.keys())] = list(clstrs[i])
    
    # Thêm thông tin về các cụm mới vào từ điển mới
    for i in clstrs2.keys():
        clstrsNew[len(clstrsNew.keys())] = list(clstrs2[i])
    
    # Tạo danh sách mới của các chỉ số
    newIdx = [j for i in clstrsNew for j in clstrsNew[i]]
    
    # Tạo ma trận tương quan mới dựa trên danh sách mới của các chỉ số
    corrNew = corr0.loc[newIdx, newIdx]
    
    # Tính toán hệ số Silhouette mới
    dist = ((1 - corr0) / 2.) ** 0.5
    kmeans_labels = np.zeros(len(dist.columns))
    for i in clstrsNew.keys():
        idxs = [dist.index.get_loc(k) for k in clstrsNew[i]]
        kmeans_labels[idxs] = i
    
    silhNew = pd.Series(silhouette_samples(dist, kmeans_labels), index=dist.index)
    
    return corrNew, clstrsNew, silhNew


''' Recursivly cluster
    Typical output: e.g if there are 4 clusters:
>>> _,_,_=clusterKMeansTop(corr0)
redo cluster:[0, 1, 2, 5]
redo cluster:[0, 1, 2]
redo cluster:[1]
redoCluster <=1:[1]
newTstatMean > tStatMean
newTstatMean > tStatMean
>>>

So it returns first time on base-case  >>>if len(redoClusters) <= 1
Then sub-sequent returnes are after the tail-recurrsion
'''
def clusterKMeansTop(corr0: pd.DataFrame, maxNumClusters=None, n_init=10):
    """
    Phân cụm các biến tài sản dựa trên ma trận tương quan sử dụng thuật toán KMeans với phương pháp tối ưu hóa.

    Args:
    corr0 (pandas.DataFrame): Ma trận tương quan gốc.
    maxNumClusters (int, optional): Số lượng cụm tối đa. Mặc định là None.
    n_init (int, optional): Số lần khởi tạo ngẫu nhiên. Mặc định là 10.

    Returns:
    pandas.DataFrame: Ma trận tương quan đã được sắp xếp lại.
    dict: Từ điển chứa thông tin về các cụm.
    pandas.Series: Series chứa hệ số Silhouette tối ưu cho từng quan sát.
    """
    if maxNumClusters is None:
        maxNumClusters = corr0.shape[1] - 1
        
    corr1, clstrs, silh = clusterKMeansBase(corr0, maxNumClusters=min(maxNumClusters, corr0.shape[1] - 1), n_init=10)
    print("clstrs length: " + str(len(clstrs.keys())))
    print("best cluster: " + str(len(clstrs.keys())))
    
    clusterTstats = {i: np.mean(silh[clstrs[i]]) / np.std(silh[clstrs[i]]) for i in clstrs.keys()}
    tStatMean = sum(clusterTstats.values()) / len(clusterTstats)
    redoClusters = [i for i in clusterTstats.keys() if clusterTstats[i] < tStatMean]
    
    if len(redoClusters) <= 2:
        print("If 2 or less clusters have a quality rating less than the average then stop.")
        print("redoCluster <= 1: " + str(redoClusters) + " clstrs len: " + str(len(clstrs.keys())))
        return corr1, clstrs, silh
    else:
        keysRedo = [j for i in redoClusters for j in clstrs[i]]
        corrTmp = corr0.loc[keysRedo, keysRedo]
        _, clstrs2, _ = clusterKMeansTop(corrTmp, maxNumClusters=min(maxNumClusters, corrTmp.shape[1] - 1), n_init=n_init)
        print("clstrs2.len, stat: " + str(len(clstrs2.keys())))
        
        # Tạo ra các đầu ra mới, nếu cần
        dict_redo_clstrs = {i: clstrs[i] for i in clstrs.keys() if i not in redoClusters}
        corrNew, clstrsNew, silhNew = makeNewOutputs(corr0, dict_redo_clstrs, clstrs2)
        newTstatMean = np.mean([np.mean(silhNew[clstrsNew[i]]) / np.std(silhNew[clstrsNew[i]]) for i in clstrsNew.keys()]) 
        
        if newTstatMean <= tStatMean:
            print("newTstatMean <= tStatMean " + str(newTstatMean) + " (len:newClst) " + str(len(clstrsNew.keys())) + " <= " + str(tStatMean) + " (len:Clst) " + str(len(clstrs.keys())))
            return corr1, clstrs, silh
        else: 
            print("newTstatMean > tStatMean " + str(newTstatMean) + " (len:newClst) " + str(len(clstrsNew.keys()))
                  + " > " + str(tStatMean) + " (len:Clst) " + str(len(clstrs.keys())))
            return corrNew, clstrsNew, silhNew